/*

 Copyright (c) 2009 Sevalecan

 This software is provided 'as-is', without any express or implied
 warranty. In no event will the authors be held liable for any damages
 arising from the use of this software.

 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it
 freely, subject to the following restrictions:

 1. The origin of this software must not be misrepresented; you must not
 claim that you wrote the original software. If you use this software
 in a product, an acknowledgment in the product documentation would be
 appreciated but is not required.

 2. Altered source versions must be plainly marked as such, and must not be
 misrepresented as being the original software.

 3. This notice may not be removed or altered from any source
 distribution.

 */

// Need to add block starting in the middle.

#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <stdlib.h>
#include <allegro.h>

namespace sevtris
{
	struct shape_part {
		int16_t width;
		int16_t height;
		const char *blockmap;
	};

	struct finalshape {
		int16_t rotation;
		shape_part rotation_part[4];
		int8_t color;
	};

	const int32_t board_width	= 10;
	const int32_t board_height	= 20;
	const int32_t block_size	= 25;
	const int32_t shape_count	= 7;
	int32_t board_x;
	int32_t board_y;
	int32_t removed_lines = 0;
	double line_speedup = 18.0;
	FONT *sevtris_font;
	BITMAP *sevtris_bg;
	DATAFILE *df;

	uint32_t centiseconds = 0;
	int32_t *colors;
	finalshape shapes[] = {
		{ 0, { {4, 1, "XXXX"}, {1, 4, "XXXX"},
				{4, 1, "XXXX"}, {1, 4, "XXXX"} }, 0 },		// stick
		{ 0, { {2, 2, "XXXX"}, {2, 2, "XXXX"},
				 {2, 2, "XXXX"}, {2, 2, "XXXX"} }, 1 },		// block
		{ 0, { {3, 2, "X..XXX"}, {2, 3, "XXX.X."},
				 {3, 2, "XXX..X"}, {2, 3, ".X.XXX"} }, 2 },	// J
		{ 0, { {3, 2, "..XXXX"}, {2, 3, "X.X.XX"},
				 {3, 2, "XXXX.."}, {2, 3, "XX.X.X"} }, 3 },	// L
		{ 0, { {3, 2, ".XXXX."}, {2, 3, "X.XX.X"},
				 {3, 2, ".XXXX."}, {2, 3, "X.XX.X"} }, 4 },	// S
		{ 0, { {3, 2, ".X.XXX"}, {2, 3, "X.XXX."},
				 {3, 2, "XXX.X."}, {2, 3, ".XXX.X"} }, 5 },	// T
		{ 0, { {3, 2, "XX..XX"}, {2, 3, ".XXXX."},
				 {3, 2, "XX..XX"}, {2, 3, ".XXXX."} }, 6 }	// Z
	};
	BITMAP *screen_buffer;
	BITMAP *board_buffer;
	BITMAP *piece_buffer;
	char **board;

	bool init();
	bool fini();
	bool run_game();
	void update_timer();
	bool collision_check(shape_part *sp, int32_t zx, int32_t zy);
	void draw_block(BITMAP *dest, int32_t x, int32_t y,
			int32_t color, int32_t block_size_ = block_size);
	void draw_shape(BITMAP *dest, finalshape &shape, int32_t x, int32_t y,
			int32_t block_size_ = block_size,
			int32_t offx = 0, int32_t offy = 0,
			int32_t rotation = -1);
	void draw_board(int32_t offx = 0, int32_t offy = 0);
	void draw_textbox(BITMAP *dest, const char *txt);
	bool remove_lines();
	bool game_over();
	bool draw_control_text(BITMAP *dest, const char *atxt, const char *btxt,
			int32_t x, int32_t y);

	inline void set_v3d_f(V3D_f &a, float b, float c, float d)
	{
		a.x = b;
		a.y = c;
		a.c = d;
	}

	inline int32_t darken_color(int32_t col, int32_t factor)
	{
		int32_t newcol;
		uint16_t r, g, b;
		r = (col >> 16) & 0xFF;
		g = (col >> 8) & 0xFF;
		b = col & 0xFF;

		if (b > factor)
			b -= factor;
		else b = 0;
		if (r > factor)
			r -= factor;
		else r = 0;
		if (g > factor)
			g -= factor;
		else g = 0;

		newcol = makecol(r, g, b);
		return newcol;
	}

	void update_timer()
	{
		++centiseconds;
	}
	END_OF_FUNCTION(update_timer);

	bool collision_check(shape_part *sp, int32_t zx, int32_t zy)
	{
		int32_t block_number;

		if ( (sp->width + zx) > board_width ||
				zx < 0 ||
				(zy + sp->height) > board_height )
			return true;
		for ( int32_t x = 0; x < sp->width; x++ )
			for ( int32_t y = 0; y < sp->height; y++ )
			{
				block_number = x + ( y * sp->width );
				if (sp->blockmap[block_number] == 'X' && board[zx+x][zy+y])
					return true;
			}

		return false;
	}

	void draw_block(BITMAP *dest, int32_t x, int32_t y,
			int32_t color, int32_t block_size_)
	{
		V3D_f vlist[4];
		set_v3d_f(vlist[0], x + 1, y + 1, color);
		set_v3d_f(vlist[1],
				x + block_size_, y + 1, darken_color(color, 60));
		set_v3d_f(vlist[2],
				x + block_size_, y + block_size_, darken_color(color, 90));
		set_v3d_f(vlist[3], x + 1, y + block_size_, darken_color(color, 60));

		quad3d_f(dest, POLYTYPE_GRGB, 0,
				&(vlist[0]), &(vlist[1]), &(vlist[2]), &(vlist[3]));

		set_v3d_f(vlist[0], x + 3, y + 3, darken_color(color, 90));
		set_v3d_f(vlist[1],
				x + block_size_ - 2, y + 3, darken_color(color, 60));
		set_v3d_f(vlist[2],
				x + block_size_ - 2, y + block_size_ - 2, color);
		set_v3d_f(vlist[3], x + 3, y + block_size_ - 2, darken_color(color, 60));

		quad3d_f(dest, POLYTYPE_GRGB, 0,
				&(vlist[0]), &(vlist[1]), &(vlist[2]), &(vlist[3]));
	}

	void draw_board(int32_t offx, int32_t offy)
	{
		for ( int32_t x = 0; x < board_width; x++ )
			for ( int32_t y = 0; y < board_height; y++ )
				if ( board[x][y] )
					draw_block(board_buffer,
							offx + x * block_size, offy + y * block_size,
							colors[board[x][y] >> 1], block_size);
	}

	void draw_shape(BITMAP *dest, finalshape &shape, int32_t x, int32_t y,
			int32_t block_size_, int32_t offx, int32_t offy, int32_t rotation)
	{
		rotation = rotation < 0 ? shape.rotation : rotation;
		int32_t block_number;
		shape_part &current_part = shape.rotation_part[rotation];

		for ( int32_t lx = 0; lx < current_part.width; lx++ )
			for ( int32_t ly = 0; ly < current_part.height; ly++ )
			{
				block_number = lx + (ly * current_part.width);
				if (current_part.blockmap[block_number] == 'X')
					draw_block(dest,
							(x + lx) * block_size + offx,
							(y + ly) * block_size + offy,
							colors[shape.color], block_size_);
			}
	}

	void draw_textbox(BITMAP *dest, const char *txt)
	{
		int32_t x, y, xoffset;

		x = dest->w/2;
		y = dest->h/2;
		xoffset = ((strlen(txt) * 8) / 2) + 2;

		rectfill(dest, x - xoffset, y - 3,
				x + xoffset, y + sevtris_font->height+2, makecol(0, 0, 0));
		rect(dest, x  -xoffset, y - 3,
				x + xoffset, y + sevtris_font->height+2, makecol(255, 255, 255));

		textout_centre_ex(dest, sevtris_font, txt, x, y, makecol(255,255,255), 0);
	}

	bool remove_lines()
	{
		bool rv = false;
		for ( int32_t y = 0; y < board_height; y++ )
		{
			int32_t line = 0;
			for ( int32_t x = 0; x < board_width; x++ )
				if ( board[x][y] )
					line++;
			if ( line == board_width )
			{
				rv = true;
				removed_lines++;
				for ( int32_t ly = y; ly > -1; ly-- )
					if ( ly > 0 )
						for ( int32_t x = 0; x < board_width; x++ )
							board[x][ly] = board[x][ly - 1];
					else
						for ( int32_t x = 0; x < board_width; x++ )
							board[x][ly] = 0;
			}
		}

		return rv;
	}

	bool draw_control_text(BITMAP *dest, const char *atxt, const char *btxt,
			int32_t x, int32_t y)
	{
		textout_ex(dest, sevtris_font, atxt, x, y,
				makecol(0,0,0), -1);
		textout_ex(dest, sevtris_font, btxt, x+60, y,
				makecol(0,0,0), -1);
	}

	bool init()
	{
			srand(time(NULL));
			allegro_init();

			install_keyboard();
			install_timer();

			set_color_depth(32);
			set_gfx_mode(GFX_AUTODETECT_WINDOWED, 800, 600, 0, 0);
		
			sevtris_font = font;

			df = load_datafile("sevtris.dat");
			sevtris_font = (FONT *)df[1].dat;
			sevtris_bg = (BITMAP *)df[0].dat;

			//sevtris_font = load_font("editundo.pcx", 0, 0);

			board_x = screen->w / 2 - board_width * block_size / 2;
			board_y = screen->h / 2 - board_height * block_size / 2;

			screen_buffer = create_bitmap(screen->w, screen->h);
			board_buffer = create_sub_bitmap(screen_buffer,
					board_x - 1, board_y - 1,
					board_width * block_size + 2,
					board_height * block_size + 2);
			piece_buffer = create_sub_bitmap(screen_buffer,
					board_x + board_width * block_size + block_size,
					board_y - 1, 4 * block_size + 3, 4 * block_size + 3);

			colors = new int32_t[shape_count];
			for ( int32_t i = 0; i < shape_count; i++ )
				colors[i] = 0;

			colors[0] = makecol(0, 200, 200);	// Stick, cyan.
			colors[1] = makecol(200, 200, 0);	// block, yellow
			colors[2] = makecol(0, 0, 200);		// J, blue.
			colors[3] = makecol(200, 130, 0);	// L, orange.
			colors[4] = makecol(0, 200, 0);		// S, Green.
			colors[5] = makecol(133, 0, 200);	// T, purple.
			colors[6] = makecol(200, 0, 0);		// Z, red.

			board = new char*[board_width];
			for ( int32_t t = 0; t < board_width; t++ )
				board[t] = new char[board_height];
			return true;
	}

	bool fini()
	{
		destroy_bitmap(piece_buffer);
		destroy_bitmap(board_buffer);
		destroy_bitmap(screen_buffer);
//		destroy_font(sevtris_font);
		unload_datafile(df);
		set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);

		for ( int32_t t = 0; t < board_width; t++ )
			delete[] board[t];
		delete[] board;
		delete[] colors;
		return true;
	}

	bool run_game()
	{
		uint32_t framecs	= 50;	// Frame in centiseconds.
		double timeunit		= 100.0;// One unit of time.
		double timediv		= 2.0;  // What to divide that unit of time by.
		bool run			= 1;	// Keep running the game loop?
		bool draw			= true;
		bool paused			= false;
		bool gameover		= false;
		uint32_t last_time	= 0;
		int32_t the_key;

		sevtris::finalshape current_shape;	// The current shape.
		bool collision = false;				// Has there been a collision?
		int16_t	next_shape = rand()%shape_count;	// Next shape.
		int32_t posx, posy;	// X and Y position of the current shape.
		posx = 0;
		posy = 0;

		for ( int32_t x = 0; x < board_width; x++ )
			for ( int32_t y = 0; y < board_height; y++ )
				board[x][y] = 0;

		install_int(sevtris::update_timer, 10);

		current_shape = sevtris::shapes[rand() % shape_count];
		posx = board_width/2 - current_shape.rotation_part[0].width/2;

		while( run )
		{
			if ( keypressed() )
			{
				the_key = readkey();
				int32_t upk = the_key & 0xFF;
				int32_t opk = the_key >> 8;
				shape_part &current_part =
					current_shape.rotation_part[current_shape.rotation];

				if ( upk == 'q' || opk == KEY_ESC )
					run = 0;
				if ( upk == 'a'  || opk == KEY_LEFT )
				{
					if ( !collision_check( &current_part, posx-1, posy) )
						posx--;
					draw = true;
				}
				if ( upk == 'd' || opk == KEY_RIGHT )
				{
					if ( !collision_check(&current_part, posx+1, posy) )
						posx++;
					draw = true;
				}
				if ( upk == 'w' || opk == KEY_UP )
				{
					int32_t nextx = posx;
					int32_t next_rotation = (current_shape.rotation + 1) % 4;
					shape_part &current_part =
						current_shape.rotation_part[current_shape.rotation];
					shape_part &next_part =
						current_shape.rotation_part[next_rotation];

					if (current_part.width+posx >= board_width )
						nextx = board_width-next_part.width;

					if ( !collision_check(&next_part, nextx, posy) )
					{
						current_shape.rotation = next_rotation;
						posx = nextx;
					}
					draw = true;
				}
				if ( upk == 'p' || opk == KEY_ENTER )
				{
					paused = !paused;
					draw = true;
				}
			}


			if ( centiseconds > (last_time + (key[KEY_S] || key[KEY_DOWN] ? 5 : framecs)) &&
					!paused && !gameover )
			{
				draw = true;
				last_time = centiseconds;
				collision = collision_check(
						&(current_shape.rotation_part[current_shape.rotation]),
						posx, posy+1
						);
				if ( !collision )
					posy++;
			}
			else if ( !draw )
				rest(1);

			if ( collision )
			{
				int8_t block_val = 1 | (current_shape.color << 1);
				shape_part &current_part =
					current_shape.rotation_part[current_shape.rotation];
				for ( int32_t x = 0; x < current_part.width; x++ )
					for ( int32_t y = 0; y < current_part.height; y++)
					{
						int32_t block_number = x + (y * current_part.width);
						if ( current_part.blockmap[block_number] == 'X' )
							board[posx+x][posy+y] = block_val;
					}
				posx = 0;
				posy = 0;
				current_shape = sevtris::shapes[next_shape];
				next_shape = rand()%shape_count;
				posx = board_width/2 - current_shape.rotation_part[0].width/2;
				collision = false;
				remove_lines();
				framecs = (int32_t)(100.0/(
								(double)removed_lines / line_speedup + 2.0 ));
				gameover = game_over();
			}


			if ( draw )
			{
				//clear_bitmap(screen_buffer);
				blit(sevtris_bg, screen_buffer, 0, 0,
						0, 0, screen_buffer->w, screen_buffer->h);

				draw_board(1, 1);
				if (!gameover)
					draw_shape(board_buffer, current_shape,
							posx, posy, block_size, 1, 1);

				rect(board_buffer,
						0, 0,
						board_buffer->w - 1, board_buffer->h - 1,
						makecol(255,255,255));

				draw_shape(piece_buffer, shapes[next_shape],
						0, 0, block_size, 1, 1);

				rect(piece_buffer, 0, 0, piece_buffer->w-1, piece_buffer->h-1,
						makecol(255,255,255));

//				textout_ex(screen_buffer, sevtris_font, "Sevtris",
//						2, 2, makecol(0,200,0), makecol(0,0,0));

				int32_t tpos = 0;
				int32_t increment = sevtris_font->height;
				int32_t loffset = 45;

				tpos += increment * 8;

				draw_control_text(screen_buffer, "Rotate", "- W, Up Arrow",
						loffset, tpos);
				tpos += increment;

				draw_control_text(screen_buffer, "Left", "- A, Left Arrow",
						loffset, tpos);
				tpos += increment;

				draw_control_text(screen_buffer, "Right", "- D, Right Arrow",
						loffset, tpos);
				tpos += increment;

				draw_control_text(screen_buffer, "Down", "- S, Down Arrow",
						loffset, tpos);
				tpos += increment;

				draw_control_text(screen_buffer, "Quit", "- Q, Escape",
						loffset, tpos);
				tpos += increment;

				draw_control_text(screen_buffer, "Pause", "- P, Enter",
						loffset, tpos);

				textprintf_right_ex(screen_buffer, sevtris_font, screen->w-20,
						2, makecol(0,0,0), -1, "Score: % 3d", removed_lines);

				if (gameover)
					draw_textbox(screen_buffer, "Game over!");
				else if (paused)
					draw_textbox(screen_buffer, "Paused");

				blit(screen_buffer, screen, 0, 0, 0, 0,
						screen_buffer->w, screen_buffer->h);

				draw = false;
			}
		};

		return true;
	}

	bool game_over()
	{
		for (int x=0;x<board_width;x++)
			if (board[x][0])
				return true;
		return false;
	}

}

int main()
{
	sevtris::init();
	sevtris::run_game();
	sevtris::fini();
	return 0;
}
END_OF_MAIN()

